// Copyright (c) 2017, teja. All rights reserved. Use of this source code
// is governed by a BSD-style license that can be found in the LICENSE file.

import 'dart:async';
import 'dart:io';
import 'dart:isolate';
import 'package:glob/glob.dart';
import 'package:vm_service_lib/vm_service_lib_io.dart' show vmServiceConnectUri;
import 'package:vm_service_lib/vm_service_lib.dart'
    show VM, VmService, IsolateRef, ReloadReport;
import 'package:watcher/watcher.dart';

const String _kVmServiceUrl = 'ws://localhost:8181/ws';

/// Hot reloader
///
///     final reloader = new HotReloader(['.']);
///     await reloader.go();
class HotReloader {
  /// The URL of the Dart VM service
  ///
  /// Default: `ws://localhost:8181/ws`.
  final String vmServiceUrl;

  /// Paths on whose changes, the application must reload
  ///
  /// It can contain [FileSystemEntity], [Uri], [String] and [Glob] only.
  /// URI's can be `package:` package url.
  final List _paths;

  final StreamController<WatchEvent> _onChange =
  new StreamController<WatchEvent>.broadcast();

  /// Fires whenever a file change. You might consider using this to trigger
  /// page reloads in a client.
  Stream<WatchEvent> get onChange => _onChange.stream;

  /// [VmService] client to request hot reloading
  VmService _client;

  StreamSubscription _sub;

  /// Initializes a hot reloader that proxies the server
  ///
  /// [paths] can contain [FileSystemEntity], [Uri], [String] and [Glob] only.
  /// URI's can be `package:` URI's as well.
  HotReloader(List paths, {this.vmServiceUrl: _kVmServiceUrl})
      : _paths = paths ?? [] {
    if (!Platform.executableArguments.contains('--observe') &&
        !Platform.executableArguments.contains('--enable-vm-service')) {
      stderr.writeln(
          'WARNING: Hot reloading requires `--enable-vm-service` or `--observe` command line flags to the Dart VM!');
      exit(1);
    }
  }

  /// Starts listening to filesystem changes
  Future go() async {
    if (_sub != null) throw new Exception('Reloader already started!');

    _sub = _onChange.stream
        .transform(new _Debounce(new Duration(seconds: 1)))
        .listen((WatchEvent event) async {
      print('Path ${event.path} changed!');
      await reload();
    });

    await _listenToFilesystem();
  }

  /// Stops listening for file system changes
  Future close() async {
    await _onChange.close();
    if (_sub != null) {
      await _sub.cancel();
      _sub = null;
    }
  }

  Future _listenToFilesystem() async {
    final paths = <String>[];

    for (var path in _paths) {
      if (path is String) {
        paths.add(path);
      } else if (path is Glob) {
        await for (final FileSystemEntity entity in path.list()) {
          paths.add(entity.path);
        }
      } else if (path is FileSystemEntity) {
        paths.add(path.path);
      } else if (path is Uri) {
        if (path.scheme == 'package') {
          final Uri uri = await Isolate.resolvePackageUri(path);
          if (uri != null) {
            paths.add(uri.toFilePath());
          } else {
            paths.add(path.toFilePath());
          }
        } else {
          paths.add(path.toFilePath());
        }
      } else {
        throw new ArgumentError('Invalid path item: $path!');
      }
    }

    for (String path in paths) {
      await _listenToStat(path);
    }
  }

  Future<bool> _listenToStat(String path) async {
    bool found = false;
    try {
      final FileStat stat = await FileStat.stat(path);
      if (stat.type == FileSystemEntityType.LINK) {
        final lnk = new Link(path);
        final String p = await lnk.resolveSymbolicLinks();
        return await _listenToStat(p);
      } else if (stat.type == FileSystemEntityType.FILE) {
        final file = new File(path);
        found = await file.exists();
      } else if (stat.type == FileSystemEntityType.DIRECTORY) {
        final dir = new Directory(path);
        found = await dir.exists();
      }

      if (!found) {
        stdout.writeln('Path: $path does not exist!');
      }

      final watcher = new Watcher(path);

      watcher.events.listen(_onChange.add, onError: (e) {
        stderr.writeln('Could not listen to file changes at ${path}: $e');
      });

      print('Listening for file changes at ${path}...');
      return true;
    } catch (e) {
      if (e is! FileSystemException) rethrow;
    }

    return false;
  }

  /// Reloads the application
  Future<Null> reload() async {
    print('Reloading server...');

    // Get vm service
    if (_client == null) _client = await vmServiceConnectUri(vmServiceUrl);

    // Find main isolate id to reload it
    final VM vm = await _client.getVM();
    final IsolateRef ref = vm.isolates.first;

    // Reload
    final ReloadReport rep = await _client.reloadSources(ref.id);

    if (!rep.success) {
      print('Reloading failed! Reason: $rep');
      exit(1);
    }
  }
}

class _Debounce<S> implements StreamTransformer<S, S> {
  final Duration _delay;

  const _Debounce(this._delay);

  Stream<S> bind(Stream<S> stream) {
    var initial = new DateTime.now();
    var next = initial.subtract(this._delay);
    return stream.where((S data) {
      var now = new DateTime.now();
      if (now.isAfter(next)) {
        next = now.add(this._delay);
        return true;
      } else {
        return false;
      }
    });
  }
}

